/*
ip_nat.c
*/

#include "inet.h"
#include "buf.h"
#include "event.h"
#include "type.h"

#include "arp.h"
#include "assert.h"
#include "clock.h"
#include "eth.h"
#include "icmp_lib.h"
#include "io.h"
#include "ip.h"
#include "ip_int.h"
#include "ipr.h"

#include "tcp.h"
#include "udp.h"

#include <net/gen/inet.h>
#include "ip_nat.h"

THIS_FILE

/* debug verbouse information */
#define DVI_I 1
#define DVI_W 2
#define DVI_R 2

PUBLIC char parse_url(ipaddr_t *dst, const char *src_str, unsigned level);
PUBLIC void ip_nat_print_hex(const void *buf, unsigned len, unsigned spline);
PUBLIC void ip_nat_init(void);
PUBLIC acc_t *ip_nat_read(ip_port_t *ip_port, acc_t *data);
PUBLIC acc_t *ip_nat_write(ip_port_t *ip_port, acc_t *data);

/* NAT working mode static parameters adjustment */
#include "ip_nat_cfg.c"

/* *************************************************************** */
/* NAT translate table */
static ip_nat_t		ip_nat[IP_NAT_ITEMS]; 
static u32_t		ip_nat_items;

static ipaddr_t		mapped_own_IPv4;
static ipaddr_t		own_IPv4;
static ip_nat_dst_t	own_IPv4r2;
static ipaddr_t		IPv4r2_src;

static u8_t			ip_opt_gc[4];

static ipaddr_t		IPv4r2_UDP_dst;
static udpport_t	IPv4r2_UDP_port; 
static ipaddr_t		IPv4r2_UDP_uni_dst;

/**/
FORWARD void mk_b45(u8_t *dst, const char *src_str);
FORWARD void mk_u12(u8_t *dst, const char *src_str);
FORWARD void mk_v6(u8_t *dst, const char *src_str);
FORWARD void mk_gc(u8_t *dst, u8_t cmd);
FORWARD void mk_init_loc(ipaddr_t *loc, const char *const *init_loc_pp );
FORWARD void mk_init_dst(ip_nat_dst_t *ip_nat_dst, init_ip_nat_dst_t *init_ip_nat_dst);
FORWARD void mk_init(ip_nat_t *ip_nat, init_ip_nat_t *init_ip_nat);

/**/
static void mk_b45( dst, src_str )
u8_t		*dst;
const char	*src_str;
#if 1
{
ipaddr_t	tmp;
u8_t		*buf;

	dst[0]= IP_OPT_R2_1; 
	dst[1]= 8; 
	memset(dst+2, 0, dst[1]-2);
	
	/* IPv4r2 set addr */
	#if 0
	/* only low byte (values d4 "0..255" accepted only) */
	dst[7]=strtoul(src_str,0,10);
	#else
	buf= (u8_t *)&tmp;

	if( parse_url(&tmp, src_str, 0) ){
		tmp= ntohl(tmp);
		dst[7]= buf[0];
		dst[6]= buf[1];
	}
	if( parse_url(&tmp, src_str, 1) ){
		tmp= ntohl(tmp);
		dst[5]= buf[0];
		dst[4]= buf[1];
		dst[3]= buf[2];
		dst[2]|= buf[3] & 0x03;
	}
	if( parse_url(&tmp, src_str, 2) ){
		tmp= ntohl(tmp);
		dst[2]|= (buf[0] & 0x07) << 2;
	}
	#endif
	
	#if DVI_I > 1
	printf("b45: ");
	ip_nat_print_hex(dst, dst[1], 16);
	#endif

	/* IPv4r2 clear flags */
	/* <flag>[3]<addr>[5] */
	dst[2] &= 0x1f;
}
#endif

static void mk_u12( dst, src_str )
u8_t		*dst;
const char	*src_str;
#if 1
{
ipaddr_t	tmp;
u8_t		*buf;

	dst[0]= IP_OPT_R2_3; 
	dst[1]= 4; 
	memset(dst+2, 0, dst[1]-2);
	
	/* IPv4r2 set addr */
	#if 0
	/* skip high halfbyte (values "0..255" accepted only) */
	dst[3]= strtoul(src_str,0,10) & 0x0ff;
	#else
	buf= (u8_t *)&tmp;

	if( parse_url(&tmp, src_str, 0) ){
		tmp= ntohl(tmp);
		dst[3]= buf[0];
		dst[2]|= buf[1] & 0x0f;
	}
	#endif
	
	#if DVI_I > 1
	printf("u12: ");
	ip_nat_print_hex(dst, dst[1], 16);
	#endif

	/* IPv4r2 clear flags, set option */
	/* <option=000>[3]<flag>[1]<high byte>[4] */
	dst[2] &= 0x0f;
	dst[2] |= 0x20;
}
#endif

/**/
static void mk_v6( dst, src_str )
u8_t		*dst;
const char	*src_str;
#if 1
{
ipaddr_t	tmp;
unsigned	n;

	dst[0]= IP_OPT_R2_1; 
	dst[1]= 20; 
	memset(dst+2, 0, dst[1]-2);
	
	/* IPv4r2 set addr */
	#if 1
	for(n=0; n<4; ++n){
	if( parse_url(&tmp, src_str, n) ){
		memcpy(
			dst+(4+n*4),
			&tmp,
			4
			);
	}}
	#endif
	
	#if DVI_I > 1
	printf("v6: ");
	ip_nat_print_hex(dst, dst[1], 16);
	#endif

	/* IPv4r2 clear flags */
	/* <flag>[1]<0>[15] */
	dst[2] &= 0x00;
	dst[3] &= 0x00;
}
#endif

/**/
static void mk_gc(u8_t *dst, u8_t cmd)
#if 1
{
	dst[0]= IP_OPT_R2_3;
	dst[1]= 3;
	dst[2]= cmd & 0x0f;

	dst[2]&= 0x0f;
	dst[2]|= 0x10;

	/* default align, is not mandatory */
	dst[3]= 0x01;
}
#endif

/**/
static void mk_init_loc(loc, init_loc_pp)
ipaddr_t	*loc;
const char	*const *init_loc_pp;
#if 1
{
	const char *init_loc= *init_loc_pp;

	if( strlen(init_loc) ){
		*loc = inet_addr( init_loc );
	}
}
#endif

static void mk_init_dst(ip_nat_dst, init_ip_nat_dst)
ip_nat_dst_t		*ip_nat_dst;
init_ip_nat_dst_t	*init_ip_nat_dst;
#if 1
{
	if( strlen(init_ip_nat_dst->ipv4) ){
		ip_nat_dst->ipv4 = inet_addr( init_ip_nat_dst->ipv4 );
	}
		
	if( strlen(init_ip_nat_dst->ip_opt_b45) ){
		ip_nat_dst->ip_opt_flags|= IP_NAT_R2_EN_B45;
		mk_b45( ip_nat_dst->b.ip_opt_b45, init_ip_nat_dst->ip_opt_b45 );
	}

	if( strlen(init_ip_nat_dst->ip_opt_u12) ){
		ip_nat_dst->ip_opt_flags|= IP_NAT_R2_EN_U12;
		mk_u12( ip_nat_dst->u.ip_opt_u12, init_ip_nat_dst->ip_opt_u12 );
	}

	if( strlen(init_ip_nat_dst->ip_opt_v6) ){
		ip_nat_dst->ip_opt_flags|= IP_NAT_R2_EN_V6;
		mk_v6( ip_nat_dst->v.ip_opt_v6, init_ip_nat_dst->ip_opt_v6 );
	}
}
#endif

static void mk_init(ip_nat, init_ip_nat)
ip_nat_t		*ip_nat;
init_ip_nat_t	*init_ip_nat;
#if 1
{
	mk_init_loc(&ip_nat->loc, &init_ip_nat->loc);
	mk_init_dst(&ip_nat->dst, &init_ip_nat->dst);
}
#endif

/* *************************************************************** */
/* 
return 0 if level not found or inet_addr failed
level 0,1,2 ... for ip0/ip1/ip2 ...
*/
char	parse_url(dst, src_str, level)
ipaddr_t	*dst;
const char	*src_str;
unsigned	level;
#if 1
{
unsigned	b, n, len, bs;
u8_t		buf[256];

u8_t		*point, *start;
unsigned	points;
ipaddr_t	tmp;
char		res;

	/* chk params */
	#if 1
	if( !dst || !src_str )return 0;

	bs= sizeof(buf);
	if(!bs)return 0;
	bs-= 1;

	len= strlen(src_str);
	if(len>bs)len= bs;
	if(!len)return 0;

	#endif

	/* find level */
	#if 1
	b=0;
	for( n=0; n<len; ++n )
	{
		if(src_str[n] != '/')continue;
		if( !(n-b) )return 0;

		if(!level)break;
		level-= 1;
		b= n+1;
	}
	if( level )return 0;
	if( b >= len )return 0;
	#endif

	/**/
	len= n-b;
	if(!len)return 0;

	memcpy(	buf, &src_str[b], len );
	buf[len]= 0;

	/*printf("MD1 buf: %s\n",buf);*/

	/* "low bytes" IPv4r2 addr format */
	#if 1
	points=0;
	for(n=0; n<len; ++n){ if( buf[n] == '.' )++points; }

	switch(points){
	case 0: 
	case 3:
	#if 1
	/*
	IPv4 accepted: 
		addr aligned to low (to dX) 
			*dst= htonl( atol(buf) ), return 1;
		and byte.byte.byte.byte
	*/
		return inet_aton((const char*)buf, dst);
	#endif

	case 1:
	case 2:
	#if 1
	/*
	aligned to low (to dX)  
		byte.byte
		byte.byte.byte
	*/
		*dst= 0;
		start= buf;
		for(n=0; n<=points; ++n){

		if(start){ point= (u8_t*)strchr((const char*)start,'.'); }
		if(point){ *point++ = 0; }

		res= inet_aton((const char*)start, &tmp);
		if(!res)return 0;
		tmp= ntohl(tmp);
		if(tmp > 255)return 0;

		/* network order 3-() */
		dst[3-(points-n)]= tmp;

		start= point;
		}
		return 1;
	#endif

	default:
		return 0;
	}
	#else
	/* IPv4 format for the level */
	return inet_aton(buf, dst);
	#endif
}
#endif

/* spline: chars per line */
void ip_nat_print_hex(buf, len, spline)
const void *buf;
unsigned len;
unsigned spline;
#if 1
{
unsigned	n,col;
u8_t		*p;

	if(!buf)return;
	if(!len)return;
	if(!spline)return;

	col= 0;
	p= buf;
	for( n=0; n<len; ++n, ++p){
		printf("0x%02x ", *p);

		col+= 1;
		if(col >= spline){
			printf("\n");
			col= 0;
		}
	}
	if(col)printf("\n");
}
#endif

/**/
PUBLIC void ip_nat_init(void)
#if 1
{
	unsigned n;

	/**/
	memset(ip_nat, 0, sizeof(ip_nat));
	ip_nat_items= sizeof(init_ip_nat)/sizeof(init_ip_nat_t);
	if( ip_nat_items > IP_NAT_ITEMS )ip_nat_items=IP_NAT_ITEMS;

	for(n=0; n<ip_nat_items; ++n)
	{
		mk_init( &ip_nat[n], &init_ip_nat[n] );

		#if DVI_I > 2
		printf("%02u: loc: ", n); writeIpAddr(ip_nat[n].loc);
		printf(" dst: ", n); writeIpAddr(ip_nat[n].dst.ipv4);
		printf("\n");
		#endif
	}

	/**/
	mapped_own_IPv4= 0;
	mk_init_loc( &mapped_own_IPv4, &init_mapped_own_IPv4 );

	/**/
	own_IPv4= 0;
	mk_init_loc( &own_IPv4, &init_own_IPv4 );
	
	/**/
	memset(&own_IPv4r2, 0, sizeof(own_IPv4r2));
	mk_init_dst( &own_IPv4r2, &init_own_IPv4r2 );

	#if DVI_I > 2
	printf("own IPv4r2: "); writeIpAddr(own_IPv4r2.ipv4);
	printf("\n");
	#endif

	/**/
	IPv4r2_src= 0;
	mk_init_loc( &IPv4r2_src, &init_IPv4r2_src );

	/**/
	memset(ip_opt_gc, 0, sizeof(ip_opt_gc));
	mk_gc(ip_opt_gc, init_IPv4r2_gate_cmd);

	/**/
	IPv4r2_UDP_dst= 0;
	mk_init_loc( &IPv4r2_UDP_dst, &init_IPv4r2_UDP_dst );

	IPv4r2_UDP_port= htons(init_IPv4r2_UDP_port);

	IPv4r2_UDP_uni_dst= 0;
	mk_init_loc( &IPv4r2_UDP_uni_dst, &init_IPv4r2_UDP_uni_dst );
}
#endif

/* must bf_afree(src_data) if return not the same */
PUBLIC acc_t *ip_nat_write(ip_port, data)
ip_port_t *ip_port;
acc_t *data;
#if 1
{
	/* vars */
	#if 1
	size_t		data_len;
	acc_t		*tmp_pack, *tmp_pack1;
	/* IP header parts: def_hdr, add_src+dst_opt, hdr_opt==prev_body, body */
	acc_t		*prev_body;

	ip_hdr_t	*ip_hdr;
	u32_t		hdr_opt_len;
	u32_t		hdr_ver, hdr_len, ip_size;

	u32_t		add_hdr_opt_tail;
	u8_t		align_opt;

	u32_t		add_src_hdr_opt_len;
	u8_t		add_src_hdr_opt[IP_R2_MTU_CTRL];

	u32_t		add_dst_hdr_opt_len;
	u8_t		add_dst_hdr_opt[IP_R2_MTU_CTRL];

	u8_t		*p;
	u32_t		n;

	ipaddr_t	old_ih_dst, old_ih_src;
	char		is_re_chksum;

	tcp_hdr_t	*tcp_hdr;
	udp_hdr_t	*udp_hdr;

	ip_nat_ps_hdr_tcp_t		
				ps_hdr_tcp;
	ip_nat_ps_hdr_udp_t		
				ps_hdr_udp;

	ip_hdr_t	*ip_hdr_udp;
	u8_t		*udp_hdr_sign;
	u32_t		ip_size_udp;
	char		is_IPv4r2_opt, is_IPv4r2_do_UDP;
	#endif

	/* check src is own IPv4 of NAT */
	if( is_own_IPv4 ){
		if( ip_port->ip_ipaddr != own_IPv4 ){
			return data;
		}
	}

	data_len= bf_bufsize(data);
	#if DVI_W > 1
	printf( "write outcomimg bs: %u\n", data_len );
	#endif

	/* check IP hdr exist */
	#if 1
	if( data_len < IP_MIN_HDR_SIZE ){
		#if DVI_W > 1
		printf("write: no ip header\n");
		#endif

		bf_afree(data);
		return 0;
	}
	#endif

	/* get IP hdr params */
	#if 1
	/* assume data already at least IP_MIN_HDR_SIZE continuous memory */
	/* but ensure we have continuous memory access to ip header */
	data= bf_packIffLess(data, IP_MIN_HDR_SIZE);

	ip_hdr= (ip_hdr_t *)ptr2acc_data(data);
	hdr_ver= (ip_hdr->ih_vers_ihl >> 4) & IH_IHL_MASK;
	hdr_len= (ip_hdr->ih_vers_ihl & IH_IHL_MASK)*4;
	ip_size= ntohs(ip_hdr->ih_length);

	old_ih_src= ip_hdr->ih_src;
	old_ih_dst= ip_hdr->ih_dst;
	is_re_chksum= 0;
	is_IPv4r2_opt= 0;
	is_IPv4r2_do_UDP= 0; 
	#endif

	/* check map own IPv4 to IPv4r2 NAT local address area */
	#if 1
	if(
		is_map_own_IPv4 
		&& ( ip_hdr->ih_dst == mapped_own_IPv4 )
	){
		/* swap src and dst addresses */
		ip_hdr->ih_dst = old_ih_src;
		ip_hdr->ih_src = old_ih_dst;

		return data;
	}
	#endif

	/* temporary check loopback here to do not change ip_write */
	#if 1
	p= (u8_t *)&ip_hdr->ih_dst;
	if( 
		( ip_hdr->ih_dst == ip_port->ip_ipaddr )
		||( p[0] == 127 )
	){ 
		return data; 
	}	
	#endif

	/* check IP frame is correct */
	#if 1
	if(
		( hdr_ver != 4 )
		||( hdr_len > ip_size )
		||( ip_size > data_len)
	){
		#if DVI_W > 1
		printf("read: wrong ip frame: " 
			   "hdr_ver: %u, hdr_len: %u, ip_size: %u\n", 
			   hdr_ver, hdr_len, ip_size);
		#endif

		bf_afree(data);
		return 0;
	}
	#endif

	/* split old ip pack to be ready to add new ip header options */
	#if 1
	tmp_pack= bf_cut(data, 0, IP_MIN_HDR_SIZE);
	/* 
	ensure we have continuous memory access to ip header 
	and empty tmp_pack->acc_next 
	*/
	tmp_pack= bf_packIffLess(tmp_pack, IP_MIN_HDR_SIZE);
	prev_body= tmp_pack;
	tmp_pack1 = tmp_pack;

	ip_hdr= (ip_hdr_t *)ptr2acc_data(tmp_pack);
	#endif

	/* split old ip header options */
	#if 1
	hdr_opt_len= hdr_len - IP_MIN_HDR_SIZE;
	if(hdr_opt_len)
	{
		tmp_pack= bf_cut(data, IP_MIN_HDR_SIZE, hdr_opt_len);
		/* 
		ensure we have continuous memory access to ip header options 
		and empty tmp_pack->acc_next 
		*/
		tmp_pack= bf_packIffLess(tmp_pack, hdr_opt_len);
		assert(!prev_body->acc_next);
		prev_body->acc_next= tmp_pack;
		prev_body= tmp_pack;
	}
	#endif

	/* split ip pack body */
	#if 1
	tmp_pack= bf_delhead(data, hdr_len);
	#if DVI_W > 1
	printf( "write body: %u, hdr: %u, ip_size: %u\n", bf_bufsize(tmp_pack), hdr_len, ip_size );
	#endif

	assert(!prev_body->acc_next);
	prev_body->acc_next= tmp_pack;
	data= tmp_pack1;
	#if DVI_W > 1
	printf( "write repacked: %u\n", bf_bufsize(data) );
	#endif
	#endif

	/* check src NAT */
	add_src_hdr_opt_len= 0;
	for(;is_own_IPv4r2;){
	#if 1
	/* 
	else assume ih_src is already correct if requested dst NAT 
	ip_hdr->ih_src == ip_nat_wan_ip;
	*/

	/* src NAT for specified IPv4 */
	if( is_IPv4r2_src ){
		if( IPv4r2_src != old_ih_src )break;
	}else{
		/* src NAT for local IPv4 */
		if( ip_port->ip_ipaddr != old_ih_src )break;
	}

	/* do src NAT */
	is_re_chksum= 1;
	is_IPv4r2_do_UDP= 1; 

	/* src IPv4 */
	ip_hdr->ih_src = own_IPv4r2.ipv4;

	/* src IPv4r2 options */
	#if 1
	if(own_IPv4r2.ip_opt_flags & IP_NAT_R2_EN_B45){
		memcpy(
			add_src_hdr_opt + add_src_hdr_opt_len, 
			own_IPv4r2.b.ip_opt_b45, 
			8);
		/* set IPv4r2 src addr extention mark */
		add_src_hdr_opt[add_src_hdr_opt_len+2] &= 0x1f;
		add_src_hdr_opt[add_src_hdr_opt_len+2] |= 0x40;
		add_src_hdr_opt_len+= 8;
	}

	if(own_IPv4r2.ip_opt_flags & IP_NAT_R2_EN_U12){
		memcpy(
			add_src_hdr_opt + add_src_hdr_opt_len, 
			own_IPv4r2.u.ip_opt_u12, 
			4);
		/* set IPv4r2 src addr extention mark */
		add_src_hdr_opt[add_src_hdr_opt_len+2] &= 0xef;
		add_src_hdr_opt[add_src_hdr_opt_len+2] |= 0x00;
		add_src_hdr_opt_len+= 4;
	}

	if(own_IPv4r2.ip_opt_flags & IP_NAT_R2_EN_V6){
		memcpy(
			add_src_hdr_opt + add_src_hdr_opt_len, 
			own_IPv4r2.v.ip_opt_v6, 
			20);
		/* set IPv4r2 src addr extention mark */
		add_src_hdr_opt[add_src_hdr_opt_len+2] &= 0x00;
		add_src_hdr_opt[add_src_hdr_opt_len+3] &= 0x00;
		add_src_hdr_opt[add_src_hdr_opt_len+2] |= 0x00;
		add_src_hdr_opt_len+= 20;
	}
	#endif
	
	break;
	/*for(;is_own_IPv4r2;)*/
	#endif
	}
	#if DVI_W > 1
	printf( "write src opts: %u\n", add_src_hdr_opt_len );
	ip_nat_print_hex(add_src_hdr_opt, add_src_hdr_opt_len, 16);
	#endif

	/* check dst NAT */
	add_dst_hdr_opt_len= 0;
	for(n=0; n<ip_nat_items; ++n){
	#if 1
	if( ip_nat[n].loc != ip_hdr->ih_dst )continue;
	
	/* do dst NAT */
	is_re_chksum= 1;
	is_IPv4r2_do_UDP= 1; 

	/* dst IPv4 */
	ip_hdr->ih_dst = ip_nat[n].dst.ipv4;

	/* dst IPv4r2 options */
	#if 1
	if(ip_nat[n].dst.ip_opt_flags & IP_NAT_R2_EN_B45){
		memcpy(
			add_dst_hdr_opt + add_dst_hdr_opt_len, 
			ip_nat[n].dst.b.ip_opt_b45, 
			8);
		/* set IPv4r2 dst addr extention mark */
		add_dst_hdr_opt[add_dst_hdr_opt_len+2] &= 0x1f;
		add_dst_hdr_opt[add_dst_hdr_opt_len+2] |= 0x80;
		add_dst_hdr_opt_len+= 8;
	}

	if(ip_nat[n].dst.ip_opt_flags & IP_NAT_R2_EN_U12){
		memcpy(
			add_dst_hdr_opt + add_dst_hdr_opt_len, 
			ip_nat[n].dst.u.ip_opt_u12, 
			4);
		/* set IPv4r2 dst addr extention mark */
		add_dst_hdr_opt[add_dst_hdr_opt_len+2] &= 0xef;
		add_dst_hdr_opt[add_dst_hdr_opt_len+2] |= 0x10;
		add_dst_hdr_opt_len+= 4;
	}

	if(ip_nat[n].dst.ip_opt_flags & IP_NAT_R2_EN_V6){
		memcpy(
			add_dst_hdr_opt + add_dst_hdr_opt_len, 
			ip_nat[n].dst.v.ip_opt_v6, 
			20);
		/* set IPv4r2 dst addr extention mark */
		add_dst_hdr_opt[add_dst_hdr_opt_len+2] &= 0x00;
		add_dst_hdr_opt[add_dst_hdr_opt_len+3] &= 0x00;
		add_dst_hdr_opt[add_dst_hdr_opt_len+2] |= 0x80;
		add_dst_hdr_opt_len+= 20;
	}
	#endif

	break;
	/*for n*/
	#endif
	}
	#if DVI_W > 1
	printf( "write dst opts: %u\n", add_dst_hdr_opt_len );
	ip_nat_print_hex(add_dst_hdr_opt, add_dst_hdr_opt_len, 16);
	#endif
	
	/* add gate cmd if IPv4 src to IPv4r2 dst */
	/* 
	in IPv4r2.1 mode IPv4r2 software avoids the IPv4 NAT
	itself emits IPv4r2 dst and does not use IPv4 servers 
	and itself generates own gate_cmd=1 if needed
	
	if IPv4 NAT in process (add_dst_hdr_opt_len exist)
	that means IPv4r2.0 mode in use	
	and IPv4 dst was generated by IPv4 software
	*/
	if(
		!add_src_hdr_opt_len
		&& add_dst_hdr_opt_len
		&& is_IPv4r2_gate_cmd
	){
	#if 1
		/**/
		memcpy(
			add_src_hdr_opt + add_src_hdr_opt_len,
			ip_opt_gc,
			ip_opt_gc[1]
			);
		add_src_hdr_opt_len += ip_opt_gc[1];

		/* align by 0x01 to 4 boundary for debug purpose */
		#if 1
		add_src_hdr_opt[add_src_hdr_opt_len]= 0x01;
		add_src_hdr_opt_len += 1;
		#endif
	
		#if DVI_W > 1
		printf( "write src+gc opts: %u\n", add_src_hdr_opt_len );
		ip_nat_print_hex(add_src_hdr_opt, add_src_hdr_opt_len, 16);
		#endif
	#endif
	}

	/* join src+dst if dst exist */
	if(add_dst_hdr_opt_len){
	#if 1
		/**/
		memcpy(
			add_src_hdr_opt + add_src_hdr_opt_len,
			add_dst_hdr_opt,
			add_dst_hdr_opt_len
			);
		add_src_hdr_opt_len += add_dst_hdr_opt_len;
	#endif
	}
	#if DVI_W > 1
	printf( "write src+dst opts: %u\n", add_src_hdr_opt_len );
	ip_nat_print_hex(add_src_hdr_opt, add_src_hdr_opt_len, 16);
	#endif

	/* add IPv4r2 options if exist */
	if (add_src_hdr_opt_len){
	#if 1

		/* 
		if there are old options align new to 4 by option 0x01 
		else align new to 4 by option 0x00
		*/
		align_opt = (hdr_opt_len)? 0x01: 0x00;

		/* ?new IPv4r2 are not aligned to 4 */
		add_hdr_opt_tail = add_src_hdr_opt_len % 4;
		if(add_hdr_opt_tail){ 
			memset( 
				add_src_hdr_opt + add_src_hdr_opt_len, 
				align_opt, 
				4-add_hdr_opt_tail 
				);
			add_src_hdr_opt_len += 4-add_hdr_opt_tail;
		}

		/* ckeck header size is less or equal IP_MAX_HDR_SIZE */
		hdr_len += add_src_hdr_opt_len;
		if( hdr_len > IP_MAX_HDR_SIZE ){
			bf_afree(data);
			return 0;
		}

		ip_size += add_src_hdr_opt_len;
		if( ip_size > IP_MAX_PACKSIZE ){
			bf_afree(data);
			return 0;
		}

		/* add new options */
		tmp_pack= bf_memreq(add_src_hdr_opt_len);
		/* ensure we have continuous memory access to */
		tmp_pack= bf_packIffLess(tmp_pack, add_src_hdr_opt_len);
		assert(!tmp_pack->acc_next);
		tmp_pack->acc_next= data->acc_next;
		data->acc_next= tmp_pack;
		
		/* move body pointer if there were no options */
		if(!hdr_opt_len)prev_body= tmp_pack;

		/**/
		memcpy(
			ptr2acc_data(tmp_pack), 
			add_src_hdr_opt, 
			add_src_hdr_opt_len
			);

		/* do correct IPv4 header */
		ip_hdr->ih_vers_ihl&= ~IH_IHL_MASK;
		ip_hdr->ih_vers_ihl|= (hdr_len/4) & IH_IHL_MASK;
		ip_hdr->ih_length= htons(ip_size);

		is_IPv4r2_opt= 1;

	#endif
	}

	/* src+dst IPv4 recount chksum */
	if(is_re_chksum){
	#if 1
	/* TCP/UDP chksum */
	switch(ip_hdr->ih_proto){
	#if 1

		/* TCP chksum */
		#if 1
		case 0x06:
		/* ensure we have continuous memory access to TCP header */
		prev_body->acc_next = bf_packIffLess(prev_body->acc_next, TCP_MIN_HDR_SIZE);

		tcp_hdr = (tcp_hdr_t*)ptr2acc_data(prev_body->acc_next);

		memset( &ps_hdr_tcp, 0, sizeof(ps_hdr_tcp) );
		ps_hdr_tcp.not_old_ih_src= ~old_ih_src;
		ps_hdr_tcp.not_old_ih_dst= ~old_ih_dst;
		ps_hdr_tcp.new_ih_src= ip_hdr->ih_src;
		ps_hdr_tcp.new_ih_dst= ip_hdr->ih_dst;

		tcp_hdr->th_chksum = ~oneC_sum(~(tcp_hdr->th_chksum), &ps_hdr_tcp, sizeof(ps_hdr_tcp));
		break;
		#endif

		/* UDP chksum */
		#if 1
		case 0x11:
		/* ensure we have continuous memory access to UDP header */
		prev_body->acc_next= bf_packIffLess(prev_body->acc_next, UDP_HDR_SIZE);

		udp_hdr = (udp_hdr_t*)ptr2acc_data(prev_body->acc_next);

		#if 1
		udp_hdr->uh_chksum = 0;

		#else
		if(udp_hdr->uh_chksum){
		
		memset( &ps_hdr_udp, 0, sizeof(ps_hdr_udp) );
		ps_hdr_udp.not_old_ih_src= ~old_ih_src;
		ps_hdr_udp.not_old_ih_dst= ~old_ih_dst;
		ps_hdr_udp.new_ih_src= ip_hdr->ih_src;
		ps_hdr_udp.new_ih_dst= ip_hdr->ih_dst;
		
		udp_hdr->uh_chksum = ~oneC_sum(~(udp_hdr->uh_chksum), &ps_hdr_udp, sizeof(ps_hdr_udp));
		}
		#endif
		break;
		#endif

	/*switch(ip_hdr->ih_proto)*/
	#endif
	}
	#endif
	}

	/* IPv4 chksum is mandatory for write */
	#if 1
	/* IPv4 header buffers must be joined for ip_write */
	tmp_pack= prev_body->acc_next;
	prev_body->acc_next= 0;
	#if DVI_W > 1
	printf( "write hdr bs: %u, hdr_len: %u\n", bf_bufsize(data), hdr_len );
	#endif
	assert( bf_bufsize(data) >= hdr_len );

	data = bf_pack(data);
	assert(!data->acc_next);
	data->acc_next= tmp_pack;
	#if DVI_W > 1
	printf( "write pack bs: %u, ip_size: %u\n", bf_bufsize(data), ip_size );
	#endif
	assert( bf_bufsize(data) >= ip_size );

	ip_hdr= (ip_hdr_t *)ptr2acc_data(data);
	ip_hdr_chksum(ip_hdr, hdr_len);
	#endif
	
	#if DVI_W > 1
	ip_nat_print_hex(ip_hdr, hdr_len, 16);
	#endif

	/* pack to IPv4 UDP for IPv4r2 */
	for(;
		is_IPv4r2_UDP 
		&& (is_IPv4r2_opt || is_IPv4r2_do_UDP)
		;
	){
	#if 1
	
	/* check min UI frame body */
	#if 1
	if( ip_size < IP_MIN_HDR_SIZE ){
		#if DVI_W > 1
		printf( "write can not make UDP frame: "
			   "no ip body: ip_size: %u\n", ip_size );
		ip_nat_print_hex(ptr2acc_data(data), ip_size, 16);
		#endif

		break;
	}
	#endif

	/**/
	ip_hdr_udp= (ip_hdr_t *)add_dst_hdr_opt;
	udp_hdr= (udp_hdr_t *)&add_dst_hdr_opt[IP_MIN_HDR_SIZE];
	udp_hdr_sign= &add_dst_hdr_opt[IP_MIN_HDR_SIZE + UDP_HDR_SIZE];

	add_dst_hdr_opt_len= IP_MIN_HDR_SIZE + UDP_HDR_SIZE + sizeof(IPv4r2_UDP_sign);
	ip_size_udp= ip_size + add_dst_hdr_opt_len;

	/* do IPv4 UDP frame: IPv4 header */
	#if 1
	ip_hdr_udp->ih_vers_ihl= 0x40;
	ip_hdr_udp->ih_vers_ihl|= (IP_MIN_HDR_SIZE/4) & IH_IHL_MASK;
	ip_hdr_udp->ih_length= htons(ip_size_udp);

	ip_hdr_udp->ih_tos= ip_hdr->ih_tos;
	ip_hdr_udp->ih_id= 0;
	ip_hdr_udp->ih_flags_fragoff= htons(IH_DONT_FRAG);
	ip_hdr_udp->ih_ttl= ip_hdr->ih_ttl;
	ip_hdr_udp->ih_proto= IPPROTO_UDP;

	ip_hdr_udp->ih_src= is_IPv4r2_UDP_own_src? ip_port->ip_ipaddr: ip_hdr->ih_src;
	ip_hdr_udp->ih_dst= is_IPv4r2_UDP_dst? IPv4r2_UDP_dst: ip_hdr->ih_dst;

	ip_hdr_chksum(ip_hdr_udp, IP_MIN_HDR_SIZE);
	#endif

	/* do IPv4 UDP frame: UDP header */
	#if 1
	udp_hdr->uh_length= htons(ip_size_udp - IP_MIN_HDR_SIZE);
	udp_hdr->uh_src_port= IPv4r2_UDP_port;
	udp_hdr->uh_dst_port= IPv4r2_UDP_port;
	udp_hdr->uh_chksum= 0;
	#endif

	/* fill IPv4 UDP frame: UDP sign */
	memcpy(
		udp_hdr_sign,
		IPv4r2_UDP_sign,
		sizeof(IPv4r2_UDP_sign)
		);

	/* add UDP frame to data */
	tmp_pack= bf_memreq(add_dst_hdr_opt_len);
	/* ensure we have continuous memory access to */
	tmp_pack= bf_packIffLess(tmp_pack, add_dst_hdr_opt_len);
	assert(!tmp_pack->acc_next);
	tmp_pack->acc_next= data;
	data= tmp_pack;

	memcpy(
		ptr2acc_data(tmp_pack), 
		add_dst_hdr_opt, 
		add_dst_hdr_opt_len
		);

	#if DVI_W > 1
	printf( "write UDP frame add_size: %u, total ip_size: %u\n", 
			add_dst_hdr_opt_len, bf_bufsize(data) );
	ip_nat_print_hex(ptr2acc_data(data), add_dst_hdr_opt_len, 16);
	#endif

	break;
	#endif
	/*for(is_IPv4r2_UDP)*/
	}

	/**/
	return data;
}
#endif

/*walk over IPv4 options*/
/* 
exmple of network performance && source code structure optimization: 
	single time IPv4 options parse
	return dst tmp_hdr_opt buf size with IPv4r2 options removed
	(return tmp_hdr_opt_len) 

to more network performance (from IPv4 up to IPv6 equ network): 
for example IPv4r2.0
first step: check IPv4r2.x options at fixed offsets 
	size= 0;
	hdr_opt[20+size] B45 src, then U12 src, then v6 src
	then B45 dst, then U12 dst, then v6 dst
	then IPv4r2 addr extention options limiter

if failed second step:
	parse options in general way
*/
PUBLIC unsigned ip_nat_walk_opt(
				/*dst IPv4 options (IPv4r2 options removed)*/
				u8_t *tmp_hdr_opt,
				/*src IPv4 options + IPv4r2 options*/
				u8_t *hdr_opt,
				u32_t hdr_opt_len,
				/*found and stored IPv4r2 options*/
				ip_nat_dst_t *src_IPv4r2, 
				ip_nat_dst_t *dst_IPv4r2
				)
#if 1
{
unsigned	i, tmp_hdr_opt_len;

char		is_addr_found;

u8_t		*hdr_src, *hdr_dst;
ip_nat_dst_t 	*ptr_IPv4r2;

	/* init ip_nat_dst_t(s) */
	#if 1
	if(src_IPv4r2)src_IPv4r2->ip_opt_flags= 0;
	if(dst_IPv4r2)dst_IPv4r2->ip_opt_flags= 0;
	#endif

	/*check input params*/
	#if 1
	if( !hdr_opt || !hdr_opt_len )return 0;
	/* zero dst is allowed with intention */
	#endif

	/* init for */
	#if 1
	hdr_src= hdr_opt;
	hdr_dst= tmp_hdr_opt;
	tmp_hdr_opt_len= 0;
	#endif

	/*walk over IPv4 options*/
	/*
	assume for options alignment free placement
	if we do full parse on read, then can 
		remove all extra 0, 1 space options
	if we do not full parse on write, then can add 
		1 to align 4 of intermediate options chunk 
		0 to align 4 of last options chunk 
	*/
	for( i=0; i<hdr_opt_len; ){
	#if 1
		/* single byte options opcodes */
		#if 1
		if( *hdr_src == 0 ){ break; }
		
		if( *hdr_src == 1 ){
			/* skip leading align if(tmp_hdr_opt_len) */
			/* skip leading align
			if(hdr_dst){ *hdr_dst++ = 1; }
			tmp_hdr_opt_len+= 1;
			*/

			i+= 1;
			/*must be last*/
			hdr_src+= 1;
			continue;
		}
		#endif

		/* invalid options */
		if( (i+2) > hdr_opt_len )break;
		if( (i+hdr_src[1]) > hdr_opt_len )break;
		if( hdr_src[1] < 2 )break;

		/* two byte options opcodes */
		switch(hdr_src[0]){

			case IP_OPT_R2_1:
			/*B45*/
			#if 1
			if( hdr_src[1] == 8 ){

			/* check IPv4r2 opcode */
			
			/* check IPv4r2 src/dst sign */
			#if	1
			is_addr_found= 0;
				/* check IPv4r2 src addr extention mark */
			if( (hdr_src[2] & 0xe0) == 0x40 ){
				ptr_IPv4r2= src_IPv4r2;
				is_addr_found= 1;
			}
				/* check IPv4r2 dst addr extention mark */
			else if( (hdr_src[2] & 0xe0) == 0x80 ){
				ptr_IPv4r2= dst_IPv4r2;
				is_addr_found= 1;
			}
			if(!is_addr_found)break;
			#endif

			/* found B45, copy to ptr_IPv4r2 */
			/* and skip invalid repeated B45 */
			#if 1
			if(ptr_IPv4r2){
			if( !(ptr_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_B45) ){

			ptr_IPv4r2->ip_opt_flags|= IP_NAT_R2_EN_B45;
			memcpy(
				ptr_IPv4r2->b.ip_opt_b45,
				hdr_src,
				hdr_src[1]
				);
			}}
			#endif

			/* remove from dst IP options*/
			i+= hdr_src[1];
			/*must be last*/
			hdr_src+= hdr_src[1];
			continue;
			}
			#endif

			/*V6*/
			#if 1
			if( hdr_src[1] == 20 ){

			/* check IPv4r2 opcode */
			if( (hdr_src[2] & 0x7f) != 0x00 )break;
			if( (hdr_src[3] & 0xff) != 0x00 )break;

			/* check IPv4r2 src/dst sign */
			#if 1
			is_addr_found= 0;
				/* check IPv4r2 src addr extention mark */
			if( (hdr_src[2] & 0x80) == 0x00 ){
				ptr_IPv4r2= src_IPv4r2;
				is_addr_found= 1;
			}
				/* check IPv4r2 dst addr extention mark */
			else if( (hdr_src[2] & 0x80) == 0x80 ){
				ptr_IPv4r2= dst_IPv4r2;
				is_addr_found= 1;
			}
			if(!is_addr_found)break;
			#endif

			/* found V6, copy to ptr_IPv4r2 */
			/* and skip invalid repeated V6 */
			#if 1
			if(ptr_IPv4r2){
			if( !(ptr_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_V6) ){

			ptr_IPv4r2->ip_opt_flags|= IP_NAT_R2_EN_V6;
			memcpy(
				ptr_IPv4r2->v.ip_opt_v6,
				hdr_src,
				hdr_src[1]
				);
			}}
			#endif

			/* remove from dst IP options*/
			i+= hdr_src[1];
			/*must be last*/
			hdr_src+= hdr_src[1];
			continue;
			}
			#endif
			break;

			case IP_OPT_R2_3:
			/*GC*/
			#if 1
			if( hdr_src[1] == 3 ){

			/* check IPv4r2 opcode */
			if( (hdr_src[2] & 0xf0) != 0x10 )break;

			/* remove gate command got for non gate host */

			/* remove from dst IP options*/
			i+= hdr_src[1];
			/*must be last*/
			hdr_src+= hdr_src[1];
			continue;
			}
			#endif

			/*U12*/
			#if 1
			if( hdr_src[1] == 4 ){

			/* check IPv4r2 opcode */
			if( (hdr_src[2] & 0xe0) != 0x20 )break;

			/* check IPv4r2 src/dst sign */
			#if 1
			is_addr_found= 0;
				/* check IPv4r2 src addr extention mark */
			if( (hdr_src[2] & 0x10) == 0x00 ){
				ptr_IPv4r2= src_IPv4r2;
				is_addr_found= 1;
			}
				/* check IPv4r2 dst addr extention mark */
			else if( (hdr_src[2] & 0x10) == 0x10 ){
				ptr_IPv4r2= dst_IPv4r2;
				is_addr_found= 1;
			}
			if(!is_addr_found)break;
			#endif

			/* found U12, copy to ptr_IPv4r2 */
			/* and skip invalid repeated U12 */
			#if 1
			if(ptr_IPv4r2){
			if( !(ptr_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_U12) ){

			ptr_IPv4r2->ip_opt_flags|= IP_NAT_R2_EN_U12;
			memcpy(
				ptr_IPv4r2->u.ip_opt_u12,
				hdr_src,
				hdr_src[1]
				);
			}}
			#endif

			/* remove from dst IP options*/
			i+= hdr_src[1];
			/*must be last*/
			hdr_src+= hdr_src[1];
			continue;
			}
			#endif
			break;
		}

		/*default: copy IP option to dst*/
		#if 1
		tmp_hdr_opt_len+= hdr_src[1];
		i+= hdr_src[1];

		if(hdr_dst){
			memcpy( 
				hdr_dst, 
				hdr_src, 
				hdr_src[1]
				);
			hdr_dst+= hdr_src[1];
		}
		/*must be last*/
		hdr_src+= hdr_src[1];
		continue;
		#endif
	
	/*for(;i<hdr_opt_len;)*/
	#endif
	}

	return tmp_hdr_opt_len;
}
#endif

/**/
PUBLIC char ip_nat_cmp_B45(
				ip_nat_dst_t *parsed_IPv4r2,
				ip_nat_dst_t *req_IPv4r2
				)
#if 1
{
	if( !parsed_IPv4r2 || !req_IPv4r2 )return 0;

	/*exist parsed B45*/
	if( parsed_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_B45 ){
	if( !(req_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_B45) )return 0;

	if( 
		(parsed_IPv4r2->b.ip_opt_b45[2] & 0x1f)
		!= (req_IPv4r2->b.ip_opt_b45[2] & 0x1f)
		)return 0;

	if( memcmp( 
			&parsed_IPv4r2->b.ip_opt_b45[3],
			&req_IPv4r2->b.ip_opt_b45[3],
			req_IPv4r2->b.ip_opt_b45[1]-3
			)
		)return 0;

	return 1;
	}

	/* not exist parsed B45*/
	if( req_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_B45 )return 0;
	return 1;
}
#endif

/**/
PUBLIC char ip_nat_cmp_U12(
				ip_nat_dst_t *parsed_IPv4r2,
				ip_nat_dst_t *req_IPv4r2
				)
#if 1
{
	if( !parsed_IPv4r2 || !req_IPv4r2 )return 0;

	/*exist parsed U12*/
	if( parsed_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_U12 ){
	if( !(req_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_U12) )return 0;

	if( 
		(parsed_IPv4r2->u.ip_opt_u12[2] & 0x0f)
		!= (req_IPv4r2->u.ip_opt_u12[2] & 0x0f)
		)return 0;

	if( memcmp( 
			&parsed_IPv4r2->u.ip_opt_u12[3],
			&req_IPv4r2->u.ip_opt_u12[3],
			req_IPv4r2->u.ip_opt_u12[1]-3
			)
		)return 0;

	return 1;
	}

	/* not exist parsed U12*/
	if( req_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_U12 )return 0;
	return 1;
}
#endif

/**/
PUBLIC char ip_nat_cmp_V6(
				ip_nat_dst_t *parsed_IPv4r2,
				ip_nat_dst_t *req_IPv4r2
				)
#if 1
{
	if( !parsed_IPv4r2 || !req_IPv4r2 )return 0;

	/*exist parsed V6*/
	if( parsed_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_V6 ){
	if( !(req_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_V6) )return 0;

	if( memcmp( 
			&parsed_IPv4r2->v.ip_opt_v6[4],
			&req_IPv4r2->v.ip_opt_v6[4],
			req_IPv4r2->v.ip_opt_v6[1]-4
			)
		)return 0;

	return 1;
	}

	/* not exist parsed V6*/
	if( req_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_V6 )return 0;
	return 1;
}
#endif

/*return hdr_opt_len*/
PUBLIC void ip_nat_buf_opt(
				/*to hdr_opt buf[hdr_opt_len]*/
				u8_t *hdr_opt,
				u32_t *hdr_opt_len,
				/*from parsed_IPv4r2 struct*/
				ip_nat_dst_t *parsed_IPv4r2
				)
#if 1
{
	if( !hdr_opt || !hdr_opt_len || !parsed_IPv4r2 )return;

	if( parsed_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_B45 ){
	memcpy(
		hdr_opt + *hdr_opt_len,
		parsed_IPv4r2->b.ip_opt_b45,
		parsed_IPv4r2->b.ip_opt_b45[1]
		);
	*hdr_opt_len+= parsed_IPv4r2->b.ip_opt_b45[1];
	}

	if( parsed_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_U12 ){
	memcpy(
		hdr_opt + *hdr_opt_len,
		parsed_IPv4r2->u.ip_opt_u12,
		parsed_IPv4r2->u.ip_opt_u12[1]
		);
	*hdr_opt_len+= parsed_IPv4r2->u.ip_opt_u12[1];
	}

	if( parsed_IPv4r2->ip_opt_flags & IP_NAT_R2_EN_V6 ){
	memcpy(
		hdr_opt + *hdr_opt_len,
		parsed_IPv4r2->v.ip_opt_v6,
		parsed_IPv4r2->v.ip_opt_v6[1]
		);
	*hdr_opt_len+= parsed_IPv4r2->v.ip_opt_v6[1];
	}

}
#endif

/**/
PUBLIC acc_t *ip_nat_read(ip_port, data)
ip_port_t *ip_port;
acc_t *data;
#if 1
{
	/*vars*/
	#if 1
	size_t		data_len;
	acc_t		*tmp_pack, *tmp_pack1;
	/* IP header parts: def_hdr, hdr_opt==prev_body, body */
	acc_t		*prev_body;

	ip_hdr_t	*ip_hdr;
	u8_t		*hdr_opt;
	u32_t		hdr_opt_len;
	u32_t		hdr_ver, hdr_len, ip_size;

	u32_t		hdr_opt_tail;
	u8_t		align_opt;

	char			is_opt_parsed;
	ip_nat_dst_t	src_IPv4r2, dst_IPv4r2;
	u8_t			tmp_hdr_opt[IP_R2_MTU_CTRL];
	u32_t			tmp_hdr_opt_len;
	u8_t			restored_hdr_opt[IP_R2_MTU_CTRL];
	u32_t			restored_hdr_opt_len;

	char		is_hdr_opt_found;
	char		is_hdr_opt_found_B45;
	char		is_hdr_opt_found_U12;
	char		is_hdr_opt_found_V6;

	char		is_src_nat_done;
	char		is_dst_nat_done;

	u32_t		n;

	u8_t		*old_hdr_opt;
	u32_t		old_hdr_opt_len;
	ipaddr_t	old_ih_dst, old_ih_src;
	char		is_re_chksum;

	tcp_hdr_t	*tcp_hdr;
	udp_hdr_t	*udp_hdr;

	ip_nat_ps_hdr_tcp_t		
				ps_hdr_tcp;
	ip_nat_ps_hdr_udp_t		
				ps_hdr_udp;
	
	u8_t		*udp_hdr_sign;
	#endif

	/* check src is own IPv4 of NAT */
	if( is_own_IPv4 ){
		if( ip_port->ip_ipaddr != own_IPv4 ){
			return data;
		}
	}

	/* unpack from IPv4 UDP for IPv4r2 */
	for(;is_IPv4r2_UDP;){
	#if 1

	data_len = bf_bufsize(data);
	#if DVI_R > 1
	printf( "read UDP incomimg bs: %u\n", data_len );
	#endif

	/* check IP hdr exist */
	#if 1
	if( data_len < IP_MIN_HDR_SIZE ){
		#if DVI_R > 1
		printf("read: no ip header\n");
		#endif

		bf_afree(data);
		return 0;
	}
	#endif

	/* split old ip pack to be ready to remove old ip header options */
	tmp_pack= bf_cut(data, 0, IP_MIN_HDR_SIZE);
	/* ensure we have continuous memory access to ip header */
	tmp_pack= bf_packIffLess(tmp_pack, IP_MIN_HDR_SIZE);

	ip_hdr= (ip_hdr_t *)ptr2acc_data(tmp_pack);
	hdr_ver= (ip_hdr->ih_vers_ihl >> 4) & IH_IHL_MASK;
	hdr_len= (ip_hdr->ih_vers_ihl & IH_IHL_MASK)*4;
	ip_size= ntohs(ip_hdr->ih_length);

	/* check IP frame is correct */
	#if 1
	if(
		( hdr_ver != 4 )
		||( hdr_len > ip_size )
		||( ip_size > data_len)
	){
		#if DVI_R > 1
		printf("read: wrong ip frame: " 
			   "hdr_ver: %u, hdr_len: %u, ip_size: %u\n", 
			   hdr_ver, hdr_len, ip_size);
		#endif

		bf_afree(tmp_pack);
		bf_afree(data);
		return 0;
	}
	#endif

	/* check 1 IP frame is IPv4r2 UI */
	#if 1
	if(
		( !is_IPv4r2_UDP_read_all )
	&&	( ip_hdr->ih_dst != ip_port->ip_ipaddr )
	&&  ( !is_IPv4r2_UDP_uni_dst
		||( ip_hdr->ih_dst != IPv4r2_UDP_uni_dst ) ) 
	){
		#if DVI_R > 1
		printf( "read UDP: drop is not IPv4r2_UDP UI: %s", "other dst: " );
		writeIpAddr(ip_hdr->ih_dst);
		printf( "\n" );
		#endif

		bf_afree(tmp_pack);
		break;
	}

	if( ip_hdr->ih_proto != IPPROTO_UDP ){
		#if DVI_R > 1
		printf( "read UDP: drop is not IPv4r2_UDP UI: %s\n", "protocol" );
		#endif

		bf_afree(tmp_pack);
		break;
	}

	/* min UI frame */
	tmp_hdr_opt_len= hdr_len 
		+ UDP_HDR_SIZE + sizeof(IPv4r2_UDP_sign) + IP_MIN_HDR_SIZE;

	if( ip_size < tmp_hdr_opt_len ){
		#if DVI_R > 1
		printf( "read UDP: drop is not IPv4r2_UDP UI: %s\n", "ip_size" );
		#endif

		bf_afree(tmp_pack);
		break;
	}
	#endif

	tmp_hdr_opt_len= UDP_HDR_SIZE + sizeof(IPv4r2_UDP_sign);

	/* split UI body */
	tmp_pack1= bf_cut(data, hdr_len, tmp_hdr_opt_len);
	/* ensure we have continuous memory access to ip header */
	tmp_pack1= bf_packIffLess(tmp_pack1, IP_MIN_HDR_SIZE);

	udp_hdr= (udp_hdr_t *)ptr2acc_data(tmp_pack1);
	udp_hdr_sign= ((u8_t *)ptr2acc_data(tmp_pack1)) + UDP_HDR_SIZE;

	/* check 2 IP frame is IPv4r2 UI */
	#if 1
	if( memcmp(
			udp_hdr_sign,
			IPv4r2_UDP_sign,
			sizeof(IPv4r2_UDP_sign)
			)
	){
		#if DVI_R > 1
		printf( "read UDP: drop is not IPv4r2_UDP UI: %s\n", "UI sign" );
		#endif

		bf_afree(tmp_pack1);
		bf_afree(tmp_pack);
		break;
	}
	#endif

	/* unpack UDP */
	bf_afree(tmp_pack1);
	bf_afree(tmp_pack);
	data= bf_delhead(data, hdr_len + tmp_hdr_opt_len);

	#if DVI_R > 1
	printf( "read UDP frame unpacked: removed head: %u\n", 
		   hdr_len + tmp_hdr_opt_len );
	ip_nat_print_hex(ptr2acc_data(data), IP_MIN_HDR_SIZE, 16);
	#endif

	break;
	#endif
	/*for(;is_IPv4r2_UDP;)*/
	}

	/**/
	data_len = bf_bufsize(data);
	#if DVI_R > 1
	printf( "read incomimg bs: %u\n", data_len );
	#endif

	/* check IP hdr exist */
	#if 1
	if( data_len < IP_MIN_HDR_SIZE ){
		#if DVI_R > 1
		printf("read: no ip header\n");
		#endif

		bf_afree(data);
		return 0;
	}
	#endif

	/* split old ip pack to be ready to remove old ip header options */
	#if 1
	tmp_pack= bf_cut(data, 0, IP_MIN_HDR_SIZE);
	/* 
	ensure we have continuous memory access to ip header 
	and empty tmp_pack->acc_next 
	*/
	tmp_pack= bf_packIffLess(tmp_pack, IP_MIN_HDR_SIZE);
	prev_body= tmp_pack;
	tmp_pack1= tmp_pack;

	ip_hdr= (ip_hdr_t *)ptr2acc_data(tmp_pack);
	hdr_ver= (ip_hdr->ih_vers_ihl >> 4) & IH_IHL_MASK;
	hdr_len= (ip_hdr->ih_vers_ihl & IH_IHL_MASK)*4;
	ip_size= ntohs(ip_hdr->ih_length);

	old_ih_src = ip_hdr->ih_src;
	old_ih_dst = ip_hdr->ih_dst;
	is_re_chksum= 0;
	#endif

	/* check IP frame is correct */
	#if 1
	if(
		( hdr_ver != 4 )
		||( hdr_len > ip_size )
		||( ip_size > data_len)
	){
		#if DVI_R > 1
		printf("read: wrong ip frame: " 
			   "hdr_ver: %u, hdr_len: %u, ip_size: %u\n", 
			   hdr_ver, hdr_len, ip_size);
		#endif

		bf_afree(tmp_pack);
		bf_afree(data);
		return 0;
	}
	#endif

	#if DVI_R > 1
	printf("read: before NAT: src: "); writeIpAddr(ip_hdr->ih_src);
	printf(" -> dst: "); writeIpAddr(ip_hdr->ih_dst);
	printf("\n");
	#endif

	/* split old ip header options */
	#if 1
	hdr_opt= 0;
	hdr_opt_len= hdr_len - IP_MIN_HDR_SIZE;

	#if DVI_R > 1
	printf("read opt_len: %u ip_size: %u\n", hdr_opt_len, ip_size);
	#endif

	if(hdr_opt_len)
	{
		tmp_pack= bf_cut(data, IP_MIN_HDR_SIZE, hdr_opt_len);
		/* 
		ensure we have continuous memory access to ip header options 
		and empty tmp_pack->acc_next 
		*/
		tmp_pack= bf_packIffLess(tmp_pack, hdr_opt_len);
		assert(!prev_body->acc_next);
		prev_body->acc_next= tmp_pack;
		prev_body= tmp_pack;

		hdr_opt= (u8_t *)ptr2acc_data(tmp_pack);

		#if DVI_R > 1
		ip_nat_print_hex(hdr_opt, hdr_opt_len, 16);
		#endif
	}

	old_hdr_opt= hdr_opt;
	old_hdr_opt_len= hdr_opt_len; 
	#endif

	/* split ip pack body */
	#if 1
	#if DVI_R > 1
	printf( "read before delhead: %u\n", bf_bufsize(data) );
	#endif
	tmp_pack= bf_delhead(data, hdr_len);
	assert(!prev_body->acc_next);
	prev_body->acc_next= tmp_pack;
	data= tmp_pack1;
	#if DVI_R > 1
	printf( "read body: %u, hdr: %u, ip_size: %u\n", bf_bufsize(tmp_pack), hdr_len, ip_size );
	#endif
	#endif

	/**/
	is_opt_parsed= 0;
	is_src_nat_done= 0;
	is_dst_nat_done= 0;

	/* check IPv4r2 dst addr */
	for(;is_own_IPv4r2;){
	#if 1
	/* 
	else assume ih_dst is already correct if requested src NAT 
	ip_hdr->ih_dst == ip_nat_wan_ip;
	*/
	
	/* check dst is own IPv4r2 of NAT */
	/* dst IPv4 */
	if( ip_hdr->ih_dst != own_IPv4r2.ipv4 )break;

	/* check dst IPv4r2 options  */
	/* parse options */
	if(!is_opt_parsed){
	#if 1
	tmp_hdr_opt_len=
		ip_nat_walk_opt(
				tmp_hdr_opt,
				hdr_opt,
				hdr_opt_len,
				&src_IPv4r2, 
				&dst_IPv4r2
				);
	is_opt_parsed= 1;
	#endif
	}

	/* compare dst addr */
	#if 1
	is_hdr_opt_found_B45= ip_nat_cmp_B45(&dst_IPv4r2, &own_IPv4r2);
	is_hdr_opt_found_U12= ip_nat_cmp_U12(&dst_IPv4r2, &own_IPv4r2);
	is_hdr_opt_found_V6= ip_nat_cmp_V6(&dst_IPv4r2, &own_IPv4r2);

	is_hdr_opt_found=
		is_hdr_opt_found_B45 
		&& is_hdr_opt_found_U12 
		&& is_hdr_opt_found_V6;
	
	#if DVI_R > 1
	printf("read dst opt found: B45: %u, U12: %u, V6: %u\n", 
		   is_hdr_opt_found_B45, is_hdr_opt_found_U12,
		   is_hdr_opt_found_V6 );
	#endif
	#if DVI_R > 1
	ip_nat_print_hex(own_IPv4r2.b.ip_opt_b45, 8, 16);
	ip_nat_print_hex(own_IPv4r2.u.ip_opt_u12, 4, 16);
	printf("flags 0x%x\n", own_IPv4r2.ip_opt_flags );

	ip_nat_print_hex(dst_IPv4r2.b.ip_opt_b45, 8, 16);
	ip_nat_print_hex(dst_IPv4r2.u.ip_opt_u12, 4, 16);
	printf("flags 0x%x, is_parsed %u\n", dst_IPv4r2.ip_opt_flags, is_opt_parsed );
	#endif
	#endif

	/**/
	if( !is_hdr_opt_found )break;
	
	/* do dst NAT */
	#if 1
	is_dst_nat_done= 1;
	is_re_chksum= 1;
	
	ip_hdr->ih_dst= (is_IPv4r2_src)? 
			/* dst NAT for specified IPv4 */
						IPv4r2_src: 
			/* dst NAT for local IPv4 */
						ip_port->ip_ipaddr;
	#if DVI_R > 1
	printf("read dst NAT loc: "); writeIpAddr(ip_hdr->ih_dst);
	printf("\n");
	#endif
	#endif

	break;
	/*for(;is_own_IPv4r2;)*/
	#endif
	}

	/* dst NAT exist but cmp failed, try IPv4r2 via IPv4 */
	if( is_own_IPv4r2 && !is_dst_nat_done ){
	#if 1
		/* check unknown IPV4r2 can not pass via the IPv4 */
		#if 1
		if( 
				/* dst NAT for specified IPv4 */
			( is_IPv4r2_src && ( ip_hdr->ih_dst == IPv4r2_src ))
				/* dst NAT for local IPv4 */
			||( !is_IPv4r2_src && ( ip_hdr->ih_dst == ip_port->ip_ipaddr ))
		){
			#if DVI_R > 1
			printf( "read: drop the unknown IPv4r2 can not pass via the IPv4: " ); 
			writeIpAddr(ip_hdr->ih_dst);
			printf( "\n" );
			ip_nat_print_hex(hdr_opt, hdr_opt_len, 16);
			#endif

			bf_afree(data);
			return 0;
		}
		#endif

		/* unknown IPV4r2 passed via the IPv4 */
		#if DVI_R > 1
		printf( "read: the unknown IPv4r2 passed via the IPv4: " ); 
		writeIpAddr(ip_hdr->ih_dst);
		printf( "\n" );
		ip_nat_print_hex(hdr_opt, hdr_opt_len, 16);
		#endif
	#endif
	}

	/* check IPv4r2 src addr */
	for(n=0; n<ip_nat_items; ++n){
	#if 1
	
	/* check IPv4 src addr */
	if( ip_nat[n].dst.ipv4 != ip_hdr->ih_src )continue;

	/* check src IPv4r2 options  */
	/* parse options */
	if(!is_opt_parsed){
	#if 1
	tmp_hdr_opt_len=
		ip_nat_walk_opt(
				tmp_hdr_opt,
				hdr_opt,
				hdr_opt_len,
				&src_IPv4r2, 
				&dst_IPv4r2
				);
	is_opt_parsed= 1;
	#endif
	}

	/* compare src addr */
	#if 1
	is_hdr_opt_found_B45= ip_nat_cmp_B45(&src_IPv4r2, &ip_nat[n].dst);
	is_hdr_opt_found_U12= ip_nat_cmp_U12(&src_IPv4r2, &ip_nat[n].dst);
	is_hdr_opt_found_V6= ip_nat_cmp_V6(&src_IPv4r2, &ip_nat[n].dst);

	is_hdr_opt_found=
		is_hdr_opt_found_B45 
		&& is_hdr_opt_found_U12 
		&& is_hdr_opt_found_V6;

	#if DVI_R > 1
	printf("read src opt found: n: %u, B45: %u, U12: %u, V6: %u\n", 
		   n, is_hdr_opt_found_B45, is_hdr_opt_found_U12,
		   is_hdr_opt_found_V6 );
	#endif
	#if DVI_R > 2
	ip_nat_print_hex(ip_nat[n].dst.b.ip_opt_b45, 8, 16);
	ip_nat_print_hex(ip_nat[n].dst.u.ip_opt_u12, 4, 16);
	printf("flags 0x%x\n", ip_nat[n].dst.ip_opt_flags);

	ip_nat_print_hex(src_IPv4r2.b.ip_opt_b45, 8, 16);
	ip_nat_print_hex(src_IPv4r2.u.ip_opt_u12, 4, 16);
	printf("flags 0x%x, is_parsed %u\n", src_IPv4r2.ip_opt_flags, is_opt_parsed);
	#endif
	#endif

	/**/
	if( !is_hdr_opt_found )continue;

	/*do src NAT*/
	#if 1
	is_src_nat_done= 1;
	is_re_chksum= 1;

	ip_hdr->ih_src = ip_nat[n].loc;

	#if DVI_R > 1
	printf("read src NAT loc: "); writeIpAddr(ip_hdr->ih_src);
	printf("\n");
	#endif
	#endif

	break;
	/*for n*/
	#endif
	}

	/* src+dst IPv4 recount UDP/TCP chksum */
	if(is_re_chksum){
	#if 1
	/* TCP/UDP chksum */
	switch(ip_hdr->ih_proto){
	#if 1
		/* TCP chksum */
		#if 1
		case 0x06:
		/* ensure we have continuous memory access to TCP header */
		prev_body->acc_next = bf_packIffLess(prev_body->acc_next, TCP_MIN_HDR_SIZE);

		tcp_hdr = (tcp_hdr_t*)ptr2acc_data(prev_body->acc_next);

		memset( &ps_hdr_tcp, 0, sizeof(ps_hdr_tcp) );
		ps_hdr_tcp.not_old_ih_src= ~old_ih_src;
		ps_hdr_tcp.not_old_ih_dst= ~old_ih_dst;
		ps_hdr_tcp.new_ih_src= ip_hdr->ih_src;
		ps_hdr_tcp.new_ih_dst= ip_hdr->ih_dst;

		tcp_hdr->th_chksum = ~oneC_sum(~(tcp_hdr->th_chksum), &ps_hdr_tcp, sizeof(ps_hdr_tcp));
		break;
		#endif

		/* UDP chksum */
		#if 1
		case 0x11:
		/* ensure we have continuous memory access to UDP header */
		prev_body->acc_next= bf_packIffLess(prev_body->acc_next, UDP_HDR_SIZE);

		udp_hdr = (udp_hdr_t*)ptr2acc_data(prev_body->acc_next);

		#if 1
		udp_hdr->uh_chksum = 0;

		#else
		if(udp_hdr->uh_chksum){
		
		memset( &ps_hdr_udp, 0, sizeof(ps_hdr_udp) );
		ps_hdr_udp.not_old_ih_src= ~old_ih_src;
		ps_hdr_udp.not_old_ih_dst= ~old_ih_dst;
		ps_hdr_udp.new_ih_src= ip_hdr->ih_src;
		ps_hdr_udp.new_ih_dst= ip_hdr->ih_dst;
		
		udp_hdr->uh_chksum = ~oneC_sum(~(udp_hdr->uh_chksum), &ps_hdr_udp, sizeof(ps_hdr_udp));
		}
		#endif
		break;
		#endif

	#endif
	/*switch(ip_hdr->ih_proto)*/
	}
	
	/*if(is_re_chksum)*/
	#endif
	}

	/* make new IPv4 options with IPv4r2 options removed */
	for(;	
		is_opt_parsed
		&& old_hdr_opt_len
		&& ( old_hdr_opt_len != tmp_hdr_opt_len )
		;
	){
	#if 1
	restored_hdr_opt_len= 0;
	
	/*check restore src opt */
	if(!is_src_nat_done){
	#if 1
		ip_nat_buf_opt(
			restored_hdr_opt,
			&restored_hdr_opt_len,
			&src_IPv4r2
			);
	#endif
	}

	/*check restore dst opt */
	if(!is_dst_nat_done){
	#if 1
		ip_nat_buf_opt(
			restored_hdr_opt,
			&restored_hdr_opt_len,
			&dst_IPv4r2
			);
	#endif
	}

	/*check align opt in tmp */
	hdr_opt_tail = (restored_hdr_opt_len + tmp_hdr_opt_len) % 4;
	if( hdr_opt_tail ){
	#if 1
		/* align tail */
		align_opt = 0x00;
		memset( 
			tmp_hdr_opt + tmp_hdr_opt_len, 
			align_opt, 
			4-hdr_opt_tail 
			);
		tmp_hdr_opt_len += 4-hdr_opt_tail;
	#endif
	}

	/* equal size options */
	/*
	can not do the skip by orig equ translated size 
	options was reparsed and realigned
	if( 
		old_hdr_opt_len 
		== ( restored_hdr_opt_len + tmp_hdr_opt_len )
	)break;
	*/

	/* create new IP options chunk */
	hdr_opt_len= restored_hdr_opt_len + tmp_hdr_opt_len; 
	#if 1
	if(!hdr_opt_len){
	/* free memory for no more options */
		assert(data->acc_next==prev_body);
		data->acc_next = prev_body->acc_next; 

		tmp_pack1= prev_body;
		prev_body= data;

	}else{
	/* create new memory */
		tmp_pack= bf_memreq(hdr_opt_len);
		tmp_pack= bf_packIffLess(tmp_pack, hdr_opt_len);
			
		assert(data->acc_next==prev_body);
		data->acc_next = tmp_pack; 
		assert(!tmp_pack->acc_next);
		tmp_pack->acc_next = prev_body->acc_next; 

		tmp_pack1= prev_body;
		prev_body= tmp_pack;

		hdr_opt= (u8_t *)ptr2acc_data(tmp_pack);
		if(restored_hdr_opt_len)
			memcpy(
				hdr_opt,
				restored_hdr_opt,
				restored_hdr_opt_len
				);

		if(tmp_hdr_opt_len)
			memcpy(
				hdr_opt + restored_hdr_opt_len,
				tmp_hdr_opt,
				tmp_hdr_opt_len
				);
	}
		
	tmp_pack1->acc_next= 0;
	bf_afree(tmp_pack1);
	#endif

	/* correct IPv4 header */
	if( hdr_opt_len != old_hdr_opt_len ){
	#if 1
		/* new IPv4 header sizes */
		hdr_len+= hdr_opt_len-old_hdr_opt_len;
		ip_size+= hdr_opt_len-old_hdr_opt_len;

		/* do correct IPv4 header */
		ip_hdr->ih_vers_ihl&= ~IH_IHL_MASK;
		ip_hdr->ih_vers_ihl|= (hdr_len/4) & IH_IHL_MASK;
		ip_hdr->ih_length= htons(ip_size);
	#endif
	}

	break;
	/*for*/
	#endif
	}

	/* IPv4 chksum is mandatory for read */
	/* join IPv4 header buffers to simplify ip_hdr_chksum */
	#if 1
	tmp_pack= prev_body->acc_next;
	prev_body->acc_next= 0;
	#if DVI_R > 1
	printf( "read hdr bs: %u, hdr_len: %u\n", bf_bufsize(data), hdr_len );
	#endif
	assert( bf_bufsize(data) >= hdr_len );

	data = bf_pack(data);
	assert(!data->acc_next);
	data->acc_next= tmp_pack;
	#if DVI_R > 1
	printf( "read pack bs: %u, ip_size: %u\n", bf_bufsize(data), ip_size );
	#endif
	assert( bf_bufsize(data) >= ip_size );
	#endif

	/* IPv4 chksum */
	ip_hdr= (ip_hdr_t *)ptr2acc_data(data);
	ip_hdr_chksum(ip_hdr, hdr_len);

	#if DVI_R > 1
	printf("read: NAT passed: src: "); writeIpAddr(ip_hdr->ih_src);
	printf(" -> dst: "); writeIpAddr(ip_hdr->ih_dst);
	printf("\n");
	#endif

	/**/
	return data;
}
#endif

/**/
